@@ -62,6 +62,18 @@
 int       p11_kit_list_modules    (int argc,
                                    char *argv[]);
 
+int       p11_kit_list_profiles   (int argc,
+                                   char *argv[]);
+
+int       p11_kit_add_profile     (int argc,
+                                   char *argv[]);
+
+int       p11_kit_delete_profile  (int argc,
+                                   char *argv[]);
+
+int       p11_kit_print_config    (int argc,
+                                   char *argv[]);
+
 int       p11_kit_trust           (int argc,
                                    char *argv[]);
 
@@ -70,6 +82,10 @@ int       p11_kit_external        (int argc,
 
 static const p11_tool_command commands[] = {
 	{ "list-modules", p11_kit_list_modules, N_("List modules and tokens") },
+	{ "list-profiles", p11_kit_list_profiles, N_("List PKCS#11 profiles supported by the token") },
+	{ "add-profile", p11_kit_add_profile, N_("Add PKCS#11 profile to the token") },
+	{ "delete-profile", p11_kit_delete_profile, N_("Delete PKCS#11 profile from the token") },
+	{ "print-config", p11_kit_print_config, N_("Print merged configuration") },
 	{ "remote", p11_kit_external, N_("Run a specific PKCS#11 module remotely") },
 	{ "server", p11_kit_external, N_("Run a server process that exposes PKCS#11 module remotely") },
 	{ P11_TOOL_FALLBACK, p11_kit_external, NULL },
@@ -85,7 +101,7 @@ p11_kit_trust (int argc,
 	args = calloc (argc + 2, sizeof (char *));
 	return_val_if_fail (args != NULL, 1);
 
-	args[0] = BINDIR "/trust";
+	args[0] = BINDIR "/trust" EXEEXT;
 	memcpy (args + 1, argv, sizeof (char *) * argc);
 	args[argc + 1] = NULL;
 
