@@ -305,7 +305,6 @@ proto_read_byte_array (p11_rpc_message *msg,
 	uint32_t length;
 	size_t vlen;
 
-	assert (len != NULL);
 	assert (msg != NULL);
 	assert (msg->input != NULL);
 
@@ -321,7 +320,8 @@ proto_read_byte_array (p11_rpc_message *msg,
 		if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &length))
 			return PARSE_ERROR;
 
-		*len = length;
+		if (len != NULL)
+			*len = length;
 
 		if (arr)
 			return CKR_BUFFER_TOO_SMALL;
@@ -333,7 +333,8 @@ proto_read_byte_array (p11_rpc_message *msg,
 	if (!p11_rpc_buffer_get_byte_array (msg->input, &msg->parsed, &val, &vlen))
 		return PARSE_ERROR;
 
-	*len = vlen;
+	if (len != NULL)
+		*len = vlen;
 
 	/* Just asking us for size */
 	if (!arr)
@@ -398,17 +399,20 @@ static void
 mechanism_list_purge (CK_MECHANISM_TYPE_PTR mechs,
                       CK_ULONG *n_mechs)
 {
-	int i;
+	CK_ULONG i;
 
 	assert (mechs != NULL);
 	assert (n_mechs != NULL);
 
-	for (i = 0; i < (int)(*n_mechs); ++i) {
-		if (!p11_rpc_mechanism_is_supported (mechs[i])) {
+	/* Trim unsupported mechanisms at the end */
+	for (; *n_mechs > 0 && !p11_rpc_mechanism_is_supported (mechs[*n_mechs - 1]); --*n_mechs)
+		;
 
+	for (i = 0; i < *n_mechs; ++i) {
+		if (!p11_rpc_mechanism_is_supported (mechs[i])) {
 			/* Remove the mechanism from the list */
 			memmove (&mechs[i], &mechs[i + 1],
-			         (*n_mechs - i) * sizeof (CK_MECHANISM_TYPE));
+				 (*n_mechs - (i + 1)) * sizeof (CK_MECHANISM_TYPE));
 
 			--(*n_mechs);
 			--i;
@@ -421,12 +425,17 @@ proto_write_mechanism (p11_rpc_message *msg,
                        CK_MECHANISM_PTR mech)
 {
 	assert (msg != NULL);
-	assert (mech != NULL);
 	assert (msg->output != NULL);
 
 	/* Make sure this is in the right order */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "M"));
 
+	/* This case is valid for C_*Init () functions to cancel operation */
+	if (mech == NULL) {
+		p11_rpc_buffer_add_uint32 (msg->output, 0);
+		return p11_buffer_failed (msg->output) ? CKR_HOST_MEMORY : CKR_OK;
+	}
+
 	if (!p11_rpc_mechanism_is_supported (mech->mechanism))
 		return CKR_MECHANISM_INVALID;
 
@@ -569,6 +578,9 @@ proto_read_sesssion_info (p11_rpc_message *msg,
 		return _ret; \
 	}
 
+#define RPC_VERSION \
+		((rpc_client *)((p11_virtual *)self)->lower_module)->version
+
 #define IN_BYTE(val) \
 	if (!p11_rpc_message_write_byte (&_msg, val)) \
 		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
@@ -577,14 +589,18 @@ proto_read_sesssion_info (p11_rpc_message *msg,
 	if (!p11_rpc_message_write_ulong (&_msg, val)) \
 		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
 
-#define IN_STRING(val) \
+#define IN_ZERO_STRING(val) \
 	if (!p11_rpc_message_write_zero_string (&_msg, val)) \
 		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
 
+#define IN_SPACE_STRING(val, len) \
+	if (!p11_rpc_message_write_space_string (&_msg, val, len)) \
+		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
+
 #define IN_BYTE_BUFFER(arr, len) \
-	if (len == NULL) \
+	if ((len) == NULL) \
 		{ _ret = CKR_ARGUMENTS_BAD; goto _cleanup; } \
-	if (!p11_rpc_message_write_byte_buffer (&_msg, arr ? *len : 0)) \
+	if (!p11_rpc_message_write_byte_buffer (&_msg, (arr) ? (*(len) > 0 ? *(len) : (uint32_t)-1) : 0)) \
 		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
 
 #define IN_BYTE_ARRAY(arr, len) \
@@ -624,8 +640,6 @@ proto_read_sesssion_info (p11_rpc_message *msg,
 		{ _ret = CKR_HOST_MEMORY; goto _cleanup; }
 
 #define IN_MECHANISM(val) \
-	if (val == NULL) \
-		{ _ret = CKR_ARGUMENTS_BAD; goto _cleanup; } \
 	_ret = proto_write_mechanism (&_msg, val); \
 	if (_ret != CKR_OK) goto _cleanup;
 
@@ -971,18 +985,46 @@ rpc_C_GetMechanismInfo (CK_X_FUNCTION_LIST *self,
 	END_CALL;
 }
 
+static CK_RV
+C_InitToken1 (CK_X_FUNCTION_LIST *self,
+              CK_SLOT_ID slot_id,
+              CK_UTF8CHAR_PTR pin, CK_ULONG pin_len,
+              CK_UTF8CHAR_PTR label)
+{
+	BEGIN_CALL_OR (C_InitToken, self, CKR_SLOT_ID_INVALID);
+		IN_ULONG (slot_id);
+		IN_BYTE_ARRAY (pin, pin_len);
+		IN_ZERO_STRING (label);
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+C_InitToken2 (CK_X_FUNCTION_LIST *self,
+              CK_SLOT_ID slot_id,
+              CK_UTF8CHAR_PTR pin, CK_ULONG pin_len,
+              CK_UTF8CHAR_PTR label)
+{
+	BEGIN_CALL_OR (C_InitToken2, self, CKR_SLOT_ID_INVALID);
+		IN_ULONG (slot_id);
+		IN_BYTE_ARRAY (pin, pin_len);
+		IN_SPACE_STRING (label, 32);
+	PROCESS_CALL;
+	END_CALL;
+}
+
 static CK_RV
 rpc_C_InitToken (CK_X_FUNCTION_LIST *self,
                  CK_SLOT_ID slot_id,
                  CK_UTF8CHAR_PTR pin, CK_ULONG pin_len,
                  CK_UTF8CHAR_PTR label)
 {
-	BEGIN_CALL_OR (C_InitToken, self, CKR_SLOT_ID_INVALID);
-		IN_ULONG (slot_id);
-		IN_BYTE_ARRAY (pin, pin_len);
-		IN_STRING (label);
-	PROCESS_CALL;
-	END_CALL;
+        uint8_t version = RPC_VERSION;
+
+        if (version == 0)
+		return C_InitToken1 (self, slot_id, pin, pin_len, label);
+	else
+		return C_InitToken2 (self, slot_id, pin, pin_len, label);
 }
 
 static CK_RV
@@ -1526,8 +1568,6 @@ rpc_C_SignUpdate (CK_X_FUNCTION_LIST *self,
                   CK_BYTE_PTR part,
                   CK_ULONG part_len)
 {
-	return_val_if_fail (part_len, CKR_ARGUMENTS_BAD);
-
 	BEGIN_CALL_OR (C_SignUpdate, self, CKR_SESSION_HANDLE_INVALID);
 		IN_ULONG (session);
 		IN_BYTE_ARRAY (part, part_len);
@@ -1879,6 +1919,370 @@ rpc_C_GenerateRandom (CK_X_FUNCTION_LIST *self,
 	END_CALL;
 }
 
+static CK_RV
+rpc_C_LoginUser (CK_X_FUNCTION_LIST *self,
+                 CK_SESSION_HANDLE session,
+                 CK_USER_TYPE user_type,
+                 CK_UTF8CHAR_PTR pin,
+                 CK_ULONG pin_len,
+                 CK_UTF8CHAR_PTR username,
+                 CK_ULONG username_len)
+{
+	BEGIN_CALL_OR (C_LoginUser, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_ULONG (user_type)
+		IN_BYTE_ARRAY (pin, pin_len)
+		IN_BYTE_ARRAY (username, username_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SessionCancel (CK_X_FUNCTION_LIST *self,
+                     CK_SESSION_HANDLE session,
+                     CK_FLAGS flags)
+{
+	BEGIN_CALL_OR (C_SessionCancel, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_ULONG (flags)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageEncryptInit (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session,
+                          CK_MECHANISM_PTR mechanism,
+                          CK_OBJECT_HANDLE key)
+{
+	BEGIN_CALL_OR (C_MessageEncryptInit, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_MECHANISM (mechanism)
+		IN_ULONG (key)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessage (CK_X_FUNCTION_LIST *self,
+                      CK_SESSION_HANDLE session,
+                      CK_VOID_PTR parameter,
+                      CK_ULONG parameter_len,
+                      CK_BYTE_PTR associated_data,
+                      CK_ULONG associated_data_len,
+                      CK_BYTE_PTR plaintext,
+                      CK_ULONG plaintext_len,
+                      CK_BYTE_PTR ciphertext,
+                      CK_ULONG_PTR ciphertext_len)
+{
+	return_val_if_fail (ciphertext_len, CKR_ARGUMENTS_BAD);
+
+	BEGIN_CALL_OR (C_EncryptMessage, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (associated_data, associated_data_len)
+		IN_BYTE_ARRAY (plaintext, plaintext_len)
+		IN_BYTE_BUFFER (ciphertext, ciphertext_len);
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY (ciphertext, ciphertext_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessageBegin (CK_X_FUNCTION_LIST *self,
+                           CK_SESSION_HANDLE session,
+                           CK_VOID_PTR parameter,
+                           CK_ULONG parameter_len,
+                           CK_BYTE_PTR associated_data,
+                           CK_ULONG associated_data_len)
+{
+	BEGIN_CALL_OR (C_EncryptMessageBegin, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (associated_data, associated_data_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_EncryptMessageNext (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session,
+                          CK_VOID_PTR parameter,
+                          CK_ULONG parameter_len,
+                          CK_BYTE_PTR plaintext_part,
+                          CK_ULONG plaintext_part_len,
+                          CK_BYTE_PTR ciphertext_part,
+                          CK_ULONG_PTR ciphertext_part_len,
+                          CK_FLAGS flags)
+{
+	BEGIN_CALL_OR (C_EncryptMessageNext, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (plaintext_part, plaintext_part_len)
+		IN_BYTE_BUFFER (ciphertext_part, ciphertext_part_len);
+		IN_ULONG (flags)
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY (ciphertext_part, ciphertext_part_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageEncryptFinal (CK_X_FUNCTION_LIST *self,
+                           CK_SESSION_HANDLE session)
+{
+	BEGIN_CALL_OR (C_MessageEncryptFinal, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageDecryptInit (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session,
+                          CK_MECHANISM_PTR mechanism,
+                          CK_OBJECT_HANDLE key)
+{
+	BEGIN_CALL_OR (C_MessageDecryptInit, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_MECHANISM (mechanism)
+		IN_ULONG (key)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessage (CK_X_FUNCTION_LIST *self,
+                      CK_SESSION_HANDLE session,
+                      CK_VOID_PTR parameter,
+                      CK_ULONG parameter_len,
+                      CK_BYTE_PTR associated_data,
+                      CK_ULONG associated_data_len,
+                      CK_BYTE_PTR ciphertext,
+                      CK_ULONG ciphertext_len,
+                      CK_BYTE_PTR plaintext,
+                      CK_ULONG_PTR plaintext_len)
+{
+	return_val_if_fail (plaintext_len, CKR_ARGUMENTS_BAD);
+
+	BEGIN_CALL_OR (C_DecryptMessage, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (associated_data, associated_data_len)
+		IN_BYTE_ARRAY (ciphertext, ciphertext_len)
+		IN_BYTE_BUFFER (plaintext, plaintext_len);
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY (plaintext, plaintext_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessageBegin (CK_X_FUNCTION_LIST *self,
+                           CK_SESSION_HANDLE session,
+                           CK_VOID_PTR parameter,
+                           CK_ULONG parameter_len,
+                           CK_BYTE_PTR associated_data,
+                           CK_ULONG associated_data_len)
+{
+	BEGIN_CALL_OR (C_DecryptMessageBegin, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (associated_data, associated_data_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_DecryptMessageNext (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session,
+                          CK_VOID_PTR parameter,
+                          CK_ULONG parameter_len,
+                          CK_BYTE_PTR ciphertext_part,
+                          CK_ULONG ciphertext_part_len,
+                          CK_BYTE_PTR plaintext_part,
+                          CK_ULONG_PTR plaintext_part_len,
+                          CK_FLAGS flags)
+{
+	BEGIN_CALL_OR (C_DecryptMessageNext, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (ciphertext_part, ciphertext_part_len)
+		IN_BYTE_BUFFER (plaintext_part, plaintext_part_len);
+		IN_ULONG (flags)
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY (plaintext_part, plaintext_part_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageDecryptFinal (CK_X_FUNCTION_LIST *self,
+                           CK_SESSION_HANDLE session)
+{
+	BEGIN_CALL_OR (C_MessageDecryptFinal, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageSignInit (CK_X_FUNCTION_LIST *self,
+                       CK_SESSION_HANDLE session,
+                       CK_MECHANISM_PTR mechanism,
+                       CK_OBJECT_HANDLE key)
+{
+	BEGIN_CALL_OR (C_MessageSignInit, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_MECHANISM (mechanism)
+		IN_ULONG (key)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessage (CK_X_FUNCTION_LIST *self,
+                   CK_SESSION_HANDLE session,
+                   CK_VOID_PTR parameter,
+                   CK_ULONG parameter_len,
+                   CK_BYTE_PTR data,
+                   CK_ULONG data_len,
+                   CK_BYTE_PTR signature,
+                   CK_ULONG_PTR signature_len)
+{
+	return_val_if_fail (signature_len, CKR_ARGUMENTS_BAD);
+
+	BEGIN_CALL_OR (C_SignMessage, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (data, data_len)
+		IN_BYTE_BUFFER (signature, signature_len);
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY (signature, signature_len)
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessageBegin (CK_X_FUNCTION_LIST *self,
+                        CK_SESSION_HANDLE session,
+                        CK_VOID_PTR parameter,
+                        CK_ULONG parameter_len)
+{
+	BEGIN_CALL_OR (C_SignMessageBegin, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_SignMessageNext (CK_X_FUNCTION_LIST *self,
+                       CK_SESSION_HANDLE session,
+                       CK_VOID_PTR parameter,
+                       CK_ULONG parameter_len,
+                       CK_BYTE_PTR data,
+                       CK_ULONG data_len,
+                       CK_BYTE_PTR signature,
+                       CK_ULONG_PTR signature_len)
+{
+        CK_ULONG null_len = 0;
+        CK_BBOOL signature_expected = signature_len ? 1 : 0;
+
+	BEGIN_CALL_OR (C_SignMessageNext, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (data, data_len)
+		IN_BYTE (signature_expected)
+		IN_BYTE_BUFFER ((signature_expected ? signature : NULL),
+				(signature_expected ? signature_len : &null_len))
+	PROCESS_CALL;
+		OUT_BYTE_ARRAY ((signature_expected ? signature : NULL),
+				(signature_expected ? signature_len : &null_len))
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageSignFinal (CK_X_FUNCTION_LIST *self,
+                        CK_SESSION_HANDLE session)
+{
+	BEGIN_CALL_OR (C_MessageSignFinal, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageVerifyInit (CK_X_FUNCTION_LIST *self,
+                         CK_SESSION_HANDLE session,
+                         CK_MECHANISM_PTR mechanism,
+                         CK_OBJECT_HANDLE key)
+{
+	BEGIN_CALL_OR (C_MessageVerifyInit, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_MECHANISM (mechanism)
+		IN_ULONG (key)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessage (CK_X_FUNCTION_LIST *self,
+                     CK_SESSION_HANDLE session,
+                     CK_VOID_PTR parameter,
+                     CK_ULONG parameter_len,
+                     CK_BYTE_PTR data,
+                     CK_ULONG data_len,
+                     CK_BYTE_PTR signature,
+                     CK_ULONG signature_len)
+{
+	BEGIN_CALL_OR (C_VerifyMessage, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (data, data_len)
+		IN_BYTE_ARRAY (signature, signature_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessageBegin (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session,
+                          CK_VOID_PTR parameter,
+                          CK_ULONG parameter_len)
+{
+	BEGIN_CALL_OR (C_VerifyMessageBegin, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_VerifyMessageNext (CK_X_FUNCTION_LIST *self,
+                         CK_SESSION_HANDLE session,
+                         CK_VOID_PTR parameter,
+                         CK_ULONG parameter_len,
+                         CK_BYTE_PTR data,
+                         CK_ULONG data_len,
+                         CK_BYTE_PTR signature,
+                         CK_ULONG signature_len)
+{
+	BEGIN_CALL_OR (C_VerifyMessageNext, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+		IN_BYTE_ARRAY (parameter, parameter_len)
+		IN_BYTE_ARRAY (data, data_len)
+		IN_BYTE_ARRAY (signature, signature_len)
+	PROCESS_CALL;
+	END_CALL;
+}
+
+static CK_RV
+rpc_C_MessageVerifyFinal (CK_X_FUNCTION_LIST *self,
+                          CK_SESSION_HANDLE session)
+{
+	BEGIN_CALL_OR (C_MessageVerifyFinal, self, CKR_SESSION_HANDLE_INVALID);
+		IN_ULONG (session)
+	PROCESS_CALL;
+	END_CALL;
+}
+
 static CK_X_FUNCTION_LIST rpc_functions = {
 	{ -1, -1 },
 	rpc_C_Initialize,
@@ -1946,6 +2350,29 @@ static CK_X_FUNCTION_LIST rpc_functions = {
 	rpc_C_SeedRandom,
 	rpc_C_GenerateRandom,
 	rpc_C_WaitForSlotEvent,
+	/* PKCS #11 3.0 */
+	rpc_C_LoginUser,
+	rpc_C_SessionCancel,
+	rpc_C_MessageEncryptInit,
+	rpc_C_EncryptMessage,
+	rpc_C_EncryptMessageBegin,
+	rpc_C_EncryptMessageNext,
+	rpc_C_MessageEncryptFinal,
+	rpc_C_MessageDecryptInit,
+	rpc_C_DecryptMessage,
+	rpc_C_DecryptMessageBegin,
+	rpc_C_DecryptMessageNext,
+	rpc_C_MessageDecryptFinal,
+	rpc_C_MessageSignInit,
+	rpc_C_SignMessage,
+	rpc_C_SignMessageBegin,
+	rpc_C_SignMessageNext,
+	rpc_C_MessageSignFinal,
+	rpc_C_MessageVerifyInit,
+	rpc_C_VerifyMessage,
+	rpc_C_VerifyMessageBegin,
+	rpc_C_VerifyMessageNext,
+	rpc_C_MessageVerifyFinal
 };
 
 static void
